package net.hostsharing.hsadminng.hs.booking.item.validators;

import net.hostsharing.hsadminng.hs.booking.item.HsBookingItem;
import net.hostsharing.hsadminng.mapper.Array;

import java.util.ArrayList;
import java.util.List;

import static net.hostsharing.hsadminng.hash.HashGenerator.Algorithm.LINUX_YESCRYPT;
import static net.hostsharing.hsadminng.hs.validation.PasswordProperty.passwordProperty;
import static net.hostsharing.hsadminng.hs.validation.StringProperty.stringProperty;

class HsDomainSetupBookingItemValidator extends HsBookingItemEntityValidator {

    public static final String FQDN_REGEX = "^((?!-)[A-Za-z0-9-]{1,63}(?<!-)\\.)+[A-Za-z]{2,12}";
    public static final String[] REGISTRAR_LEVEL_DOMAINS = Array.of(
            // "[^.]+", // top-level-domains are already rejected by FQDN_REGEX
            "(co|org|gov|ac|sch)\\.uk",
            "(com|net|org|edu|gov|asn|id)\\.au",
            "(co|ne|or|ac|go)\\.jp",
            "(com|net|org|gov|edu|ac)\\.cn",
            "(com|net|org|gov|edu|mil|art)\\.br",
            "(co|net|org|gen|firm|ind)\\.in",
            "(com|net|org|gob|edu)\\.mx",
            "(gov|edu)\\.it",
            "(co|net|org|govt|ac|school|geek|kiwi)\\.nz",
            "(co|ne|or|go|re|pe)\\.kr"
    );
    public static final String DOMAIN_NAME_PROPERTY_NAME = "domainName";
    public static final String VERIFICATION_PASSPHRASE_PROPERTY_NAME = "verificationPassphrase";

    HsDomainSetupBookingItemValidator() {
        super(
                stringProperty(DOMAIN_NAME_PROPERTY_NAME).writeOnce()
                        .matchesRegEx(FQDN_REGEX).describedAs("is not a (non-top-level) fully qualified domain name")
                        .notMatchesRegEx(REGISTRAR_LEVEL_DOMAINS).describedAs("is a forbidden registrar-level domain name")
                        .required(),
                passwordProperty(VERIFICATION_PASSPHRASE_PROPERTY_NAME).minLength(8).maxLength(64)
                        .hashedUsing(LINUX_YESCRYPT).writeOnly().optional()
        );
    }

    @Override
    public List<String> validateEntity(final HsBookingItem bookingItem) {
        final var violations = new ArrayList<String>();
        final var domainName = bookingItem.getDirectValue(DOMAIN_NAME_PROPERTY_NAME, String.class);
        if ( !bookingItem.isLoaded() &&
                domainName.matches("hostsharing.(com|net|org|coop)") ) {
            violations.add("'" + bookingItem.toShortString() + ".resources." + DOMAIN_NAME_PROPERTY_NAME + "' = '" + domainName + "' is a forbidden Hostsharing domain name");
        }
        violations.addAll(super.validateEntity(bookingItem));
        return violations;
    }
}
