package net.hostsharing.hsadminng.hs.hosting.asset.validators;

import org.apache.commons.collections4.EnumerationUtils;

import javax.naming.InvalidNameException;
import javax.naming.NameNotFoundException;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.ServiceUnavailableException;
import javax.naming.directory.Attribute;
import javax.naming.directory.InitialDirContext;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;

import static java.util.Arrays.stream;
import static java.util.Collections.emptyList;

public class Dns {

    private final static Map<String, Result> fakeResults = new HashMap<>();

    public static void fakeResultForDomain(final String domainName, final Result fakeResult) {
        fakeResults.put(domainName, fakeResult);
    }

    static void resetFakeResults() {
        fakeResults.clear();
    }

    public enum Status {
        SUCCESS,
        NAME_NOT_FOUND,
        INVALID_NAME,
        SERVICE_UNAVAILABLE,
        UNKNOWN_FAILURE
    }

    public record Result(Status status, List<String> records, NamingException exception) {

        public static Result fromRecords(final NamingEnumeration<?> recordEnumeration) {
            final List<String> records = recordEnumeration == null
                    ? emptyList()
                    : EnumerationUtils.toList(recordEnumeration).stream().map(Object::toString).toList();
            return new Result(Status.SUCCESS, records, null);
        }

        public static Result fromRecords(final String... records) {
            return new Result(Status.SUCCESS, stream(records).toList(), null);
        }

        public static Result fromException(final NamingException exception) {
            return switch (exception) {
                case ServiceUnavailableException exc -> new Result(Status.SERVICE_UNAVAILABLE, null, exc);
                case NameNotFoundException exc -> new Result(Status.NAME_NOT_FOUND, null, exc);
                case InvalidNameException exc -> new Result(Status.INVALID_NAME, null, exc);
                case NamingException exc -> new Result(Status.UNKNOWN_FAILURE, null, exc);
            };
        }
    }

    private final String domainName;

    public Dns(final String domainName) {
        this.domainName = domainName;
    }

    public Result fetchRecordsOfType(final String recordType) {
        if (fakeResults.containsKey(domainName)) {
            return fakeResults.get(domainName);
        }

        try {
            final var env = new Hashtable<>();
            env.put("java.naming.factory.initial", "com.sun.jndi.dns.DnsContextFactory");
            final Attribute records = new InitialDirContext(env)
                    .getAttributes(domainName, new String[] { recordType })
                    .get(recordType);
            return Result.fromRecords(records != null ? records.getAll() : null);
        } catch (final NamingException exception) {
            return Result.fromException(exception);
        }
    }

    public static void main(String[] args) {
        final var result = new Dns("example.org").fetchRecordsOfType("TXT");
        System.out.println(result);
    }

}
