package net.hostsharing.hsadminng.hs.hosting.asset.validators;

import net.hostsharing.hsadminng.hs.booking.item.HsBookingItemRealEntity;
import net.hostsharing.hsadminng.hs.booking.item.HsBookingItemType;
import net.hostsharing.hsadminng.hs.booking.item.validators.HsBookingItemEntityValidatorRegistry;
import net.hostsharing.hsadminng.hs.hosting.asset.HsHostingAssetRbacEntity;
import net.hostsharing.hsadminng.hs.hosting.asset.HsHostingAssetRealEntity;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.junit.jupiter.params.provider.ValueSource;

import javax.naming.InvalidNameException;
import javax.naming.NameNotFoundException;
import javax.naming.NamingException;
import javax.naming.ServiceUnavailableException;
import java.util.HashMap;
import java.util.Map;

import static java.util.Map.entry;
import static java.util.Map.ofEntries;
import static net.hostsharing.hsadminng.hs.hosting.asset.HsHostingAssetType.CLOUD_SERVER;
import static net.hostsharing.hsadminng.hs.hosting.asset.HsHostingAssetType.DOMAIN_SETUP;
import static net.hostsharing.hsadminng.hs.hosting.asset.HsHostingAssetType.MANAGED_SERVER;
import static org.assertj.core.api.Assertions.assertThat;

class HsDomainSetupHostingAssetValidatorUnitTest {

    static HsHostingAssetRbacEntity.HsHostingAssetRbacEntityBuilder<?, ?> validEntityBuilder(final String domainName) {
        final HsBookingItemRealEntity bookingItem = HsBookingItemRealEntity.builder()
                .type(HsBookingItemType.DOMAIN_SETUP)
                .resources(new HashMap<>(ofEntries(
                        entry("domainName", domainName),
                        entry("verificationPassphrase", "some secret verification passphrase")
                )))
                .build();
        HsBookingItemEntityValidatorRegistry.forType(HsBookingItemType.DOMAIN_SETUP).prepareProperties(null, bookingItem);
        return HsHostingAssetRbacEntity.builder()
                .type(DOMAIN_SETUP)
                .bookingItem(bookingItem)
                .identifier(domainName);
    }

    static HsHostingAssetRbacEntity.HsHostingAssetRbacEntityBuilder<?, ?> validEntityBuilder() {
        return validEntityBuilder("example.org");
    }

    @AfterEach
    void cleanup() {
        Dns.resetFakeResults();
    }

    enum InvalidDomainNameIdentifier {
        EMPTY(""),
        TOO_LONG("abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyz0123456890123456789.de"),
        DASH_AT_BEGINNING("-example.com"),
        DOT_AT_BEGINNING(".example.com"),
        DOT_AT_END("example.com.");

        final String domainName;

        InvalidDomainNameIdentifier(final String domainName) {
            this.domainName = domainName;
        }
    }

    @ParameterizedTest
    @EnumSource(InvalidDomainNameIdentifier.class)
    void rejectsInvalidIdentifier(final InvalidDomainNameIdentifier testCase) {
        // given
        final var givenEntity = validEntityBuilder().identifier(testCase.domainName).build();
        final var validator = HostingAssetEntityValidatorRegistry.forType(givenEntity.getType());

        // when
        final var result = validator.validateEntity(givenEntity);

        // then
        assertThat(result).contains(
                "'identifier' expected to match 'example.org', but is '"+testCase.domainName+"'"
        );
    }

    enum ValidDomainNameIdentifier {
        SIMPLE("example.org"),
        MAX_LENGTH("abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyz01234568901.de"),
        WITH_DASH("example-test.com"),
        SUBDOMAIN("test.example.com");

        final String domainName;

        ValidDomainNameIdentifier(final String domainName) {
            this.domainName = domainName;
        }
    }

    @ParameterizedTest
    @EnumSource(ValidDomainNameIdentifier.class)
    void acceptsValidIdentifier(final ValidDomainNameIdentifier testCase) {
        // given
        Dns.fakeResultForDomain(testCase.domainName, new Dns.Result(Dns.Status.NAME_NOT_FOUND, null, null));
        final var givenEntity = validEntityBuilder(testCase.domainName).identifier(testCase.domainName).build();
        final var validator = HostingAssetEntityValidatorRegistry.forType(givenEntity.getType());

        // when
        final var result = validator.validateEntity(givenEntity);

        // then
        assertThat(result).isEmpty();
    }

    @Test
    void containsNoProperties() {
        // when
        final var validator = HostingAssetEntityValidatorRegistry.forType(CLOUD_SERVER);

        // then
        assertThat(validator.properties()).map(Map::toString).isEmpty();
    }

    @Test
    void validatesReferencedEntities() {
        // given
        final var domainSetupHostingAssetEntity = validEntityBuilder()
                .parentAsset(HsHostingAssetRealEntity.builder().type(CLOUD_SERVER).build())
                .assignedToAsset(HsHostingAssetRealEntity.builder().type(MANAGED_SERVER).build())
                .bookingItem(HsBookingItemRealEntity.builder().type(HsBookingItemType.CLOUD_SERVER).build())
                .build();
        final var validator = HostingAssetEntityValidatorRegistry.forType(domainSetupHostingAssetEntity.getType());

        // when
        final var result = validator.validateEntity(domainSetupHostingAssetEntity);

        // then
        assertThat(result).containsExactlyInAnyOrder(
                "'DOMAIN_SETUP:example.org.bookingItem' or parentItem must be null but is of type CLOUD_SERVER",
                "'DOMAIN_SETUP:example.org.parentAsset' must be null or of type DOMAIN_SETUP but is of type CLOUD_SERVER",
                "'DOMAIN_SETUP:example.org.assignedToAsset' must be null but is of type MANAGED_SERVER");
    }

    @Test
    void rejectsDomainNameNotMatchingBookingItemDomainName() {
        // given
        final var domainSetupHostingAssetEntity = validEntityBuilder()
                .identifier("not-matching-booking-item-domain-name.org")
                .build();
        final var validator = HostingAssetEntityValidatorRegistry.forType(domainSetupHostingAssetEntity.getType());

        // when
        final var result = validator.validateEntity(domainSetupHostingAssetEntity);

        // then
        assertThat(result).containsExactlyInAnyOrder(
                "'identifier' expected to match 'example.org', but is 'not-matching-booking-item-domain-name.org'");
    }

    @ParameterizedTest
    @ValueSource(strings = {"not-matching-booking-item-domain-name.org", "indirect.subdomain.example.org"})
    void rejectsDomainNameWhichIsNotADirectSubdomainOfParentAsset(final String newDomainName) {
        // given
        final var domainSetupHostingAssetEntity = validEntityBuilder()
                .bookingItem(null)
                .parentAsset(createValidParentDomainSetupAsset("example.org"))
                .identifier(newDomainName)
                .build();
        final var validator = HostingAssetEntityValidatorRegistry.forType(domainSetupHostingAssetEntity.getType());

        // when
        final var result = validator.validateEntity(domainSetupHostingAssetEntity);

        // then
        assertThat(result).containsExactlyInAnyOrder(
                "'identifier' expected to match '(\\*|(?!-)[A-Za-z0-9-]{1,63}(?<!-))\\.example\\.org', " +
                        "but is '" + newDomainName + "'");
    }

    @Test
    void expectsEitherParentAssetOrBookingItem() {

        // given
        final var domainSetupHostingAssetEntity = validEntityBuilder().build();
        Dns.fakeResultForDomain(
                domainSetupHostingAssetEntity.getIdentifier(),
                new Dns.Result(Dns.Status.NAME_NOT_FOUND, null, null));
        final var validator = HostingAssetEntityValidatorRegistry.forType(domainSetupHostingAssetEntity.getType());

        // when
        final var result = validator.validateEntity(domainSetupHostingAssetEntity);

        // then
        assertThat(result).isEmpty();
    }

    enum DnsLookupFailureTestCase {
        SERVICE_UNAVAILABLE(
                new ServiceUnavailableException("no Internet connection"),
                "[DNS] lookup failed for domain name 'example.org': javax.naming.ServiceUnavailableException: no Internet connection"),
        NAME_NOT_FOUND(
                new NameNotFoundException("domain name not found"),
                null), // no
        INVALID_NAME(
                new InvalidNameException("domain name too long or whatever"),
                "[DNS] invalid domain name 'example.org'"),
        UNKNOWN_FAILURE(
                new NamingException("some other problem"),
                "[DNS] lookup failed for domain name 'example.org': javax.naming.NamingException: some other problem");

        public final NamingException givenException;
        public final String expectedErrorMessage;

        DnsLookupFailureTestCase(final NamingException givenException, final String expectedErrorMessage) {
            this.givenException = givenException;
            this.expectedErrorMessage = expectedErrorMessage;
        }
    }

    @ParameterizedTest
    @EnumSource(DnsLookupFailureTestCase.class)
    void handlesDnsLookupFailures(final DnsLookupFailureTestCase testCase) {

        // given
        final var domainSetupHostingAssetEntity = validEntityBuilder().build();
        Dns.fakeResultForDomain(
                domainSetupHostingAssetEntity.getIdentifier(),
                Dns.Result.fromException(testCase.givenException));
        final var validator = HostingAssetEntityValidatorRegistry.forType(domainSetupHostingAssetEntity.getType());

        // when
        final var result = validator.validateEntity(domainSetupHostingAssetEntity);

        // then
        if (testCase.expectedErrorMessage != null) {
            assertThat(result).containsExactly(testCase.expectedErrorMessage);
        } else {
            assertThat(result).isEmpty();
        }
    }

    @Test
    void allowSetupOfNonExistingSubdomainOfRegistrarLevelDomain() {

        // given
        final var domainSetupHostingAssetEntity = validEntityBuilder().build();
        final var domainName = domainSetupHostingAssetEntity.getIdentifier();
        Dns.fakeResultForDomain(
                domainName,
                Dns.Result.fromException(new NameNotFoundException("domain not registered")));
        final var validator = HostingAssetEntityValidatorRegistry.forType(domainSetupHostingAssetEntity.getType());

        // when
        final var result = validator.validateEntity(domainSetupHostingAssetEntity);

        // then
        assertThat(result).isEmpty();
    }

    @Test
    void rejectSetupOfExistingDomainWithInvalidDnsVerification() {

        // given
        final var domainSetupHostingAssetEntity = validEntityBuilder().build();
        final var domainName = domainSetupHostingAssetEntity.getIdentifier();
        Dns.fakeResultForDomain(
                domainName,
                Dns.Result.fromRecords("Hostsharing-domain-setup-verification-code=SOME-DEFINITELY-WRONG-HASH"));
        final var validator = HostingAssetEntityValidatorRegistry.forType(domainSetupHostingAssetEntity.getType());

        // when
        final var result = validator.validateEntity(domainSetupHostingAssetEntity);

        // then
        assertThat(result).contains("[DNS] no TXT record 'Hostsharing-domain-setup-verification=...' with valid hash found for domain name 'example.org'");
    }

    @Test
    void allowSetupOfExistingDomainWithValidDnsVerification() {

        // given
        final var domainSetupHostingAssetEntity = validEntityBuilder().build();
        final var domainName = domainSetupHostingAssetEntity.getIdentifier();
        final var expectedHash = domainSetupHostingAssetEntity.getBookingItem().getDirectValue("verificationPassphrase", String.class);
        Dns.fakeResultForDomain(
                domainName,
                Dns.Result.fromRecords("Hostsharing-domain-setup-verification-code=" + expectedHash));
        final var validator = HostingAssetEntityValidatorRegistry.forType(domainSetupHostingAssetEntity.getType());

        // when
        final var result = validator.validateEntity(domainSetupHostingAssetEntity);

        // then
        assertThat(result).isEmpty();
    }

    private static HsHostingAssetRealEntity createValidParentDomainSetupAsset(final String parentDomainName) {
        final var bookingItem = HsBookingItemRealEntity.builder()
                .type(HsBookingItemType.DOMAIN_SETUP)
                .resources(ofEntries(
                        entry("domainName", parentDomainName)
                ))
                .build();
        final var parentAsset = HsHostingAssetRealEntity.builder()
                .type(DOMAIN_SETUP)
                .bookingItem(bookingItem)
                .identifier(parentDomainName).build();
        return parentAsset;
    }
}
